from __future__ import annotations
from typing import Union

import numpy as np

from numpy import typing as nptyping

import xarray as xr
import torch



NDArray = Union[nptyping.NDArray[np.float_], nptyping.NDArray[np.int_], nptyping.NDArray[np.bool_], xr.DataArray, torch.Tensor]

number = Union[int, float, complex, np.int_, np.float_, np.complex_, np.bool_]


# class NumberArray(Protocol):
#
#     def __array__(self, dtype: Any=...) -> ndarray[Any, Any]: ...
#
#     def __neg__(self) -> NumberArray: ...
#
#     def __abs__(self) -> NumberArray: ...
#
#     def __add__(self, other: _ArrayLike) -> NumberArray: ...
#
#     def __radd__(self, other: _ArrayLike) -> NumberArray: ...
#
#     def __mul__(self, other: _ArrayLike) -> NumberArray: ...
#     def __rmul__(self, other: _ArrayLike) -> NumberArray: ...
#
#     def __truediv__(self, other: _ArrayLike) -> NumberArray: ...
#     def __rtruediv__(self, other: _ArrayLike) -> NumberArray: ...
#
#     def __floordiv__(self, other: _ArrayLike) -> NumberArray: ...
#     def __rfloordiv__(self, other: _ArrayLike) -> NumberArray: ...
#
#
#     def __mod__(self, other: _ArrayLike) -> NumberArray: ...
#     def __rmod__(self, other: _ArrayLike) -> NumberArray: ...
#
#     def __and__(self, other: _ArrayLike) -> BoolArray: ...
#     def __rand__(self, other: _ArrayLike) -> BoolArray: ...
#
#     def __or__(self, other: _ArrayLike) -> BoolArray: ...
#     def __ror__(self, other: _ArrayLike) -> BoolArray: ...
#
#     def __gt__(self, other: _ArrayLike) -> BoolArray: ...
#     def __ge__(self, other: _ArrayLike) -> BoolArray: ...
#
#     def __lt__(self, other: _ArrayLike) -> BoolArray: ...
#     def __le__(self, other: _ArrayLike) -> BoolArray: ...
#    
#     def __eq__(self, other: _ArrayLike) -> BoolArray: ...
#
#
# class BoolArray(Protocol):
#
#     def __array__(self, dtype: Any) -> ndarray[Any, Any]: ...
#
#     def __and__(self, other: _ArrayLike) -> BoolArray: ...
#     def __rand__(self, other: _ArrayLike) -> BoolArray: ...
#
#     def __or__(self, other: _ArrayLike) -> BoolArray: ...
#     def __ror__(self, other: _ArrayLike) -> BoolArray: ...
#
#     def __eq__(self, other: _ArrayLike) -> BoolArray: ...
#    
#
#



from __future__ import annotations

from typing import Any, List, Protocol, Tuple, cast, Union
import numpy as np

import xarray as xr
from xarray import DataArray
from xtensors.typing import NDArray, number, DimLike

from ._broadcast import broadcast_arrays

from xtensors.unify import get_axis_from_dim, strip_dims, get_axes

from scipy import special


def where(condition: NDArray, x: NDArray|number, y: NDArray|number) -> DataArray:
    '''
        The returned tensor is named only if condition is named
    '''
    dims = None

    X = cast(Union[NDArray,np.number], x)
    Y = cast(Union[NDArray,np.number], y)

    _c = condition.__array__()
    _x = X
    _y = Y

    if hasattr(x, 'shape'):
        _x = cast(NDArray, X.__array__())
        _c, _x, _, _ = broadcast_arrays(_c, _x)

    if hasattr(y, 'shape'):
        _y = cast(NDArray, Y.__array__())
        _c, _y, _, _ = broadcast_arrays(_c, _y)

    _z = np.where(_c, _x, _y)

    if isinstance(condition, DataArray): 
        dims = list(condition.dims)
        if len(_z.shape) > len(dims):
            dims = ['dim_i' for i in range(len(_z.shape)-len(dims))] + dims

    return DataArray(_z, dims=dims)


def softmax(x: NDArray, dim: DimLike) -> DataArray:

    axis = get_axes(x, dim)[0]
    _y = special.softmax(x.__array__(), axis=axis)

    dims = None
    if isinstance(x, DataArray):
        dims = list(x.dims)

    return DataArray(_y, dims=dims)



def get_rank(x: NDArray|list|number) -> int:
    if hasattr(x, 'shape'):
        return len(cast(NDArray, x).shape)

    if isinstance(x, list):
        return get_rank(x[0]) + 1

    return 0





from __future__ import annotations
from typing import Any, Protocol, Sequence, Tuple, overload, cast
import numpy as np
from xarray import DataArray

from xtensors.typing import NDArray
from xtensors.typing import AxisDimPair, DimLike, DimsLike

from xtensors.unify import get_axes, strip_dims

'''
For reduction functions that can act over multiple axes/dims.
'''


class _np_reduction_func(Protocol):
    def __call__(self, a: NDArray, axis: int|Tuple[int,...]) -> DataArray: ...


class ReductionFunc(Protocol):
    def __call__(self, x: NDArray, dim: DimLike|DimsLike|None) -> DataArray: ...




def _reduction_factory(_np_func: _np_reduction_func) -> ReductionFunc:
    def _reduce(x: NDArray, dim: DimLike|DimsLike|None=None) -> DataArray:

        _x = x.__array__()
        _axis = get_axes(x, dim)

        newdims = None
        if isinstance(x, DataArray):
            olddims = cast(Tuple[str,...], x.dims)
            newdims = strip_dims(olddims, _axis)

        _y = _np_func(_x, axis=_axis)

        return DataArray(_y, dims=newdims)
    return _reduce


_sum = _reduction_factory(np.sum)
_mean = _reduction_factory(np.mean)
_std = _reduction_factory(np.std)

_nanmean = _reduction_factory(np.nanmean)
_nanstd = _reduction_factory(np.nanstd)
_nansum = _reduction_factory(np.nansum)

_max = _reduction_factory(np.max)
_min = _reduction_factory(np.min)

_nanmax = _reduction_factory(np.nanmax)
_nanmin = _reduction_factory(np.nanmin)

_all = _reduction_factory(np.all)
_any = _reduction_factory(np.any)






from __future__ import annotations
from typing import cast, overload, Tuple, Protocol, Any

import numpy as np

from xarray import DataArray

from xtensors.typing import NDArray
from xtensors.typing import DimLike, DimsLike

from xtensors.unify import get_axes, strip_dims



def diagonal(x: NDArray, dim1: DimLike, dim2: DimLike, dim_out: str) -> DataArray:

    assert isinstance(dim1, (str, int, tuple))
    _x = x.__array__()

    axes1 =  get_axes(x, dim=dim1)
    axes2 =  get_axes(x, dim=dim2)

    assert len(axes1) == len(axes2) == 1

    axis1 = axes1[0]
    axis2 = axes2[0]

    y = np.diagonal(_x, axis1=axis1, axis2=axis2)

    if isinstance(x, DataArray):
        olddims = cast(Tuple[str], x.dims)
        newdims = list(strip_dims(olddims, (axis1, axis2)))
        newdims.append(dim_out)
        return DataArray(y, dims=newdims)
    else:
        y_ = DataArray(y)
        y_ = y_.rename({y_.dims[-1]: dim_out})
        return y_






from __future__ import annotations
import typing


from xarray import DataArray
from xtensors.typing import NDArray


class Functional:
    def __init__(self):
        self.name : str = 'NOTIMPLEMENTED_FUNCTIONAL'

    def __call__(self, x: NDArray) -> DataArray:
        raise NotImplementedError


class Pipe(Functional):
    '''
        Pipe(f1, f2, ..., fn)(x) = fn(...f2(f1(x))...)
    '''
    def __init__(self, *f: Functional, delim: str='.'):
        self.f = f
        self.delim = delim
        self.name = delim.join([_f.name for _f in self.f])

    def __call__(self, x: NDArray) -> DataArray:
        _y = x
        for _f in self.f:
            _y = _f(_y)

        if not isinstance(_y, DataArray):
            _y = DataArray(_y)
        return _y
